package trip.model;

import java.util.Arrays;
import java.util.HashSet;

public class DataValidator {
	
	private static final HashSet<String> reservedNonType = new HashSet<String>(
			Arrays.asList(new String[]{"abstract", "default", "if", 
					"private", "break", "import", "public", 
					"this", "do", "implements", "protected", 
					"throws", "else", "instanceof", "return", 
					"transient", "case", "extends", "try", 
					"catch", "final", "interface", "static",
					"finally", "strictfp", "volatile", "class", 
					"native", "super", "while", "const", "for", 
					"new", "switch", "continue", "goto", "package", 
					"synchronized", "throw", "null", "true", "false"}));
	
	private static final HashSet<String> reservedTypes = new HashSet<String>(
			Arrays.asList(new String[]{"double", "boolean", "byte", "int", "short",
									   "void", "char", "long", "float"}));
	

	public final static boolean isValidPackageName(String pkg){
		boolean isValid = true;
		String[] packages = pkg.split("\\.");
		if (packages.length == 0){
			return false;
		}
		// Check first and last character (boundary condition for that split())
		if(!Character.isJavaIdentifierStart(pkg.charAt(0)) 
				|| !Character.isJavaIdentifierPart(pkg.charAt(pkg.length()-1)) ){
			return false;
		}
		// Check for repeated periods (not caught by split stuff either)
		if(pkg.contains("..")){
			return false;
		}
		for(String part : packages){
			if (!isValidJavaIdentifier(part)){
				isValid = false;
				break;
			}
		}
		return isValid;
	}	
	
	/**
	 * Check whether the given String is a valid Java type name. Some reserved words
	 * are valid types, such as int, void, etc.
	 * @param typeName The type name to check.
	 * @return True is the string is a valid type name, otherwise false.
	 */
	public final static boolean isValidJavaTypeName(String typeName) {
		if (!isValidIdentifier(typeName)) return false;
		if (isReservedNonType(typeName)) return false;
		return true;
	}
	
	/**
	 * Check whether the given String is a valid Java identifier.
	 *
	 * @param ident String to check
	 *
	 * @return True if the given String is a valid Java
	 *         identifier, false otherwise.
	 */
	public final static boolean isValidJavaIdentifier(String ident) {
		if (!isValidIdentifier(ident)) return false;
		if (isReservedWord(ident)) return false;
		return true;
	}
	
	
	private static boolean isValidIdentifier(String string) {
		if (string == null || string.length() == 0) {
			return false;
		}
		char[] charIdent = string.toCharArray();
		if (!Character.isJavaIdentifierStart(charIdent[0])) {
			return false;
		}
		for (int i = 1; i < charIdent.length; i++) {
			if (!Character.isJavaIdentifierPart(charIdent[i])) {
				return false;
			}
		}
		return true;
	}
	
	private static boolean isReservedWord(String string) {
		if(reservedNonType.contains(string)) {
			return true;
		}
		if (reservedTypes.contains(string)) {
			return true;
		}
		return false;
	}
	
	private static boolean isReservedNonType(String string) {
		if (reservedNonType.contains(string)) {
			return true;
		}
		return false;
	}
	
	/**
	 * Check whether the given String has already been used as the name of
	 * a parameter in the given String array.
	 *
	 * @param parameterName String to check
	 * @param usedNames Array of strings already used as parameter names
	 * @return True if the given String has not been used, false otherwise.
	 */
	public static boolean isValidParameter(String parameterName, String[] usedNames) {
		for(String parameter : usedNames) {
			if(getParamName(parameter).equals(parameterName)) {
				return false;
			}
		}
		return true;
	}

	private static String getParamName(String paramStr) {
		if (paramStr.startsWith("final ")) 
			paramStr = paramStr.substring(paramStr.indexOf(' ')+1);
		paramStr = paramStr.substring(paramStr.indexOf(' ')+1);
		return paramStr;
	}
}
