package trip.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * An abstract Java component. This class defines components of Java code
 * at the most abstract level.
 *
 */
public abstract class JavaComponent {
	
	private String name;
	private JavaCompositeComponent parent;
	private RelationshipStrategy relationshipStrategy;
	
	/**
	 * <p>Add a non-null child Java component to this components collection of children.
	 * Subclasses are responsible for determining if the given component is
	 * a valid child, and should throw an IllegalArgumentException if not.</p>
	 * 
	 * Outside clients should only change the structure of this tree by using setParent().
	 * 
	 * @param component The child component to add to this component.
	 * @return True if the child was successfully added to the component,
	 *          otherwise false.
	 * @throws IllegalArgumentException if the client tries to add an invalid
	 *         component to this component, for example if the client tries to
	 *         add a Method component to a Package. Also throws this exception if
	 *         the client attempts to add a null child.
	 */
	protected abstract boolean addChild(JavaComponent component) throws IllegalArgumentException;
	/**
	 * Remove a component from this components collection of children.
	 * @param component The component to remove.
	 * @return True if the component was in the collection of children and was successfully
	 *         removed, false otherwise.
	 */
	protected abstract boolean removeChild(JavaComponent component);
	/**
	 * Gets an <b>unmodifiable</b> collection of this component's child components. 
	 * @return A collection of child components. If the component has no child
	 *         components then an empty unmodifiable collection is returned.
	 */
	public abstract Collection<JavaComponent> getChildren();
	/**
	 * Gets this component's parent component, or null if it has not been set.
	 * @return This component's parent, or null if this component is an orphan.
	 */
	public JavaCompositeComponent getParent() {
		return parent;
	}
	/**
	 * <p>Sets the parent of this component.</p> 
	 * <p><b>Side Effects:</b> Calling this method results in this component adding itself 
	 * to the new parent's collection of children. If it can't successfully do this then
	 * the method returns false.</p>
	 * <p>If newParent is null then this component is removed from its current parent
	 * and becomes an orphaned component.</p>
	 * 
	 * @param newParent The new parent of this component. Null to orphan this component.
	 */
	public boolean setParent(JavaCompositeComponent newParent) {
		if (newParent != null) {
			// Can't be your own parent
			if  (newParent == this) return false;
			
			// Ensure we don't create a cyclic graph
			Collection<JavaComponent> extendedChildren = getExtendedChildren();
			if (extendedChildren.contains(newParent)) return false;
		}
		
		
		try {
			if (newParent != null) newParent.addChild(this);
			// We do the remove after the newParent add because the add may throw an exception.
			// Doing the remove first could result in an orphaned component.
			if (parent != null) parent.removeChild(this);
			this.parent = newParent;
			return true;
		} catch (IllegalArgumentException ex) {
			// log error
			return false;
		}
	}
	
	/**
	 * <p>Convenience method for getting all children of this of this component, and all
	 * children of all of its children.</p>
	 */
	protected Collection<JavaComponent> getExtendedChildren() {
		List<JavaComponent> allChildren = new ArrayList<JavaComponent>();
		if (getChildren() == null) return allChildren;
		for (JavaComponent jc : getChildren()) {
			allChildren.add(jc);
			allChildren.addAll(jc.getExtendedChildren());
		}
		return allChildren;
	}
	
	/**
	 * <p>Gets the fully qualified name of the JavaComponent. This name will be in the form of
	 * <code>package.type.member</code> where package may be constructed of several dot-separated
	 * packages.</p>
	 * 
	 * @return The fully qualified name of this component.
	 */
	public String getQualifiedName() {
		if (getName() == null) return null;
		if (getParent() == null ||
			getParent().getQualifiedName() == null ||
			getParent().getQualifiedName().equals("")) return getName();
		return getParent().getQualifiedName() + "." + getName();
	}
	
	protected String getName() { return null; }
	public void setName(String newName) { name = newName; }
	

	public Set<JavaRelationship> resolveRelationships() {
		if (relationshipStrategy == null) return new HashSet<JavaRelationship>();
		return relationshipStrategy.resolveRelationships();
	}
	public void setRelationshipStrategy(RelationshipStrategy strategy) {
		relationshipStrategy = strategy;
	}
}
