package trip.model;

/**
 * Represents a package in Java. Packages are composed of other packages and of types.
 * Packages can be parented by other packages or by projects.
 *
 */
public class JavaPackage extends JavaCompositeComponent {

	public static final String DEFAULT_PACKAGE_STRING = "(default)";
	
	private String name;
	
	public JavaPackage(String name) {
		if (name.indexOf('.') > -1) throw new 
			IllegalArgumentException("Package names must not contain dots. "+
					"Construct multiple packages or use convenience methods");
		this.name = name;
	}
	
	/**
	 * Adds a new child component to the JavaPackage. JavaPackages can be composed of:
	 * <ul>
	 * <li>JavaPackage</li>
	 * <li>JavaType
	 *   <ul>
	 *   <li>JavaInterface</li>
	 *   <li>JavaClass</li>
	 *   <li>JavaEnum</li>
	 *   </ul></li>
	 * </ul>
	 * 
	 * <p>Outside clients are expected to use setParent() on JavaComponents to change the structure
	 * of the tree. This method will not complete a two-way dependency between a JavaPackage and its
	 * new child, it will only add the child to its child collection.</p>
	 * <p>Duplicate children will be ignored (return false).</p> 
	 * 
	 * @throws IllegalArgumentException if the given child is not a valid child
	 *                                  for a JavaPackage.
	 */
	@Override
	protected boolean addChild(JavaComponent component) throws IllegalArgumentException {
		if (component instanceof JavaPackage ||
			component instanceof JavaType)
			return super.addChild(component);
		else throw new IllegalArgumentException
			("JavaPackages can only be composed of JavaTypes or JavaPackages");
	}
	
	/**
	 * Sets the parent of this JavaPackage and also sets this JavaPackage as a child of the parameter
	 * object. JavaPackages can be parented by
	 * <ul><li>JavaPackage</li>
	 * <li>JavaProject</li></ul> 
	 * <p><b>Side Effects:</b> Calling this method results in this component adding itself 
	 * to the new parent's collection of children. If it can't successfully do this then
	 * the method returns false.</p>
	 * <p>If newParent is null then this component is removed from its current parent
	 * and becomes an orphaned component.</p>
	 * 
	 * @param newParent The new parent of this component. Null to orphan this component.
	 * @return True if the parent was changed, otherwise false.
	 */
	@Override
	public boolean setParent(JavaCompositeComponent newParent) {
		if (newParent == null || 
			newParent instanceof JavaPackage)
			return super.setParent(newParent);
		else throw new IllegalArgumentException
			("JavaPackages can only be children of JavaProjects or JavaPackages.");
	}
	
	/**
	 * <p>Creates a new package from a package String. If the String represents nested packages
	 * (for example, pkg1.pkg2.pkg3) then the resulting JavaPackage will be correctly nested. The
	 * resulting object reference will point to the most highly nested package (in the previous
	 * example, pk3).</p>
	 * @param packageString
	 * @return
	 */
	public static JavaPackage fromString(String packageString) throws IllegalArgumentException {
		// Usual guard code
		if (packageString == null || packageString.trim().equals(""))
			throw new IllegalArgumentException("Package String cannot be null or empty.");
		if (!DataValidator.isValidPackageName(packageString) &&
			!packageString.equals(DEFAULT_PACKAGE_STRING))
			throw new IllegalArgumentException("Package name is malformed");
		String[] identifiers = packageString.split("\\.");
		JavaPackage prevPkg = null;
		for (int i = 0; i < identifiers.length; i++) {
			
			JavaPackage pkg = new JavaPackage(identifiers[i]);
			pkg.setParent(prevPkg);
			prevPkg = pkg;
		}
		return prevPkg;
	}
	
	
	public String getName() {
		if (name.equals(DEFAULT_PACKAGE_STRING)) return "";
		return name;
	}
}
