package trip.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * This class represents a type in Java. Types are composed of methods, fields
 * and other types.
 *
 */
public abstract class JavaType extends JavaCompositeComponent {
	
	private String name;
	private List<JavaInterface> superInterfaces;
	

	public JavaType(String name) {
		this.name = name;
		superInterfaces = new ArrayList<JavaInterface>();
	}

	/**
	 * Adds a new child component to the JavaType. JavaTypes can be composed of:
	 * <ul>
	 * <li>JavaType
	 *   <ul>
	 *   <li>JavaInterface</li>
	 *   <li>JavaClass</li>
	 *   <li>JavaEnum</li>
	 *   </ul></li>
	 * <li>JavaMember
	 *   <ul>
	 *   <li>JavaField</li>
	 *   <li>JavaMethod</li>
	 *   </ul></li>
	 * </ul>
	 * 
	 * <p>Outside clients are expected to use setParent() on JavaComponents to change the structure
	 * of the tree. This method will not complete a two-way dependency between a JavaType and its
	 * new child, it will only add the child to its child collection.</p>
	 * <p>Duplicate children will be ignored (return false).</p> 
	 * 
	 * @return True if the child was successfully added, otherwise false.
	 * @throws IllegalArgumentException if the given child is not a valid child
	 *                                  for a JavaType.
	 */
	@Override
	protected boolean addChild(JavaComponent component) throws IllegalArgumentException {
		if (component instanceof JavaType ||
			component instanceof JavaMember) 
			return super.addChild(component);
		else throw new IllegalArgumentException
			 ("JavaTypes can only be composed of JavaTypes or JavaMembers.");
	}	
	
	/**
	 * Sets the parent of this JavaType and also sets this JavaType as a child of the parameter
	 * object. JavaTypes can be parented by
	 * <ul><li>JavaType</li>
	 * <li>JavaPackage</li></ul> 
	 * <p><b>Side Effects:</b> Calling this method results in this component adding itself 
	 * to the new parent's collection of children. If it can't successfully do this then
	 * the method returns false.</p>
	 * <p>If newParent is null then this component is removed from its current parent
	 * and becomes an orphaned component.</p>
	 * 
	 * @param newParent The new parent of this component. Null to orphan this component.
	 * @return True if the parent was changed, otherwise false.
	 */
	@Override
	public boolean setParent(JavaCompositeComponent newParent) {
		if (newParent == null || 
			newParent instanceof JavaType ||
			newParent instanceof JavaPackage)
			return super.setParent(newParent);
		else throw new IllegalArgumentException
			("JavaTypes can only be children of JavaTypes or JavaPackages.");
	}
	
	public String getName() {
		return name;
	}
	
	/**
	 * Gets an <strong>unmodifiable</strong> list of super interfaces. In the case
	 * of a JavaClass this is the list of interfaces that this class implements. In
	 * the case of a JavaInterface it is the list of interfaces that this interface
	 * extends.
	 * @return An unmodifiable list of extended/implmented interfaces.
	 */
	public List<JavaInterface> getSuperInterfaces() {
		return Collections.unmodifiableList(superInterfaces);
	}
	public boolean addSuperInterface(JavaInterface interfac) {
		if (superInterfaces.contains(interfac)) return false;
		return superInterfaces.add(interfac);
	}
	
	/**
	 * <p>This constructs a new JavaClass from a fully qualified name, parented by newly constructed
	 * packages. For example, if the client passes in the String <code>pkg1.pkg2.Class</code> then the
	 * resulting JavaClass will be a class of the name <code>Class</code>, parented by a JavaPackage
	 * <code>pkg2</code>, which is parented by a JavaPackage <code>pkg1</code>.</p>
	 * 
	 * <p>Names without qualifiers will return an unparented JavaClass of that name (equivalent to using
	 * the named constructor). Invalid names will throw an unchecked IllegalArgumentException.</p>
	 * 
	 * @param qualifiedName The fully qualified name of the type you wish to construct.
	 * @return The type, parented appropriately.
	 */
	public static JavaClass classFromQualifiedName(String qualifiedName) {
		if (!DataValidator.isValidPackageName(qualifiedName))
			throw new IllegalArgumentException("Qualified name must be valid Java identifier to construct type.");
		if (qualifiedName.indexOf(".") < 0) return new JavaClass(qualifiedName);
		JavaPackage pkg = JavaPackage.fromString(qualifiedName.substring(0, qualifiedName.lastIndexOf(".")));
		JavaClass result = new JavaClass(qualifiedName.substring(qualifiedName.lastIndexOf(".")+1));
		result.setParent(pkg);
		
		return result;
	}
	/**
	 * <p>This constructs a new JavaInterface from a fully qualified name, parented by newly constructed
	 * packages. For example, if the client passes in the String <code>pkg1.pkg2.Interface</code> then the
	 * resulting JavaInterface will be an interface of the name <code>Interface</code>, parented by a JavaPackage
	 * <code>pkg2</code>, which is parented by a JavaPackage <code>pkg1</code>.</p>
	 * 
	 * <p>Names without qualifiers will return an unparented JavaInterface of that name (equivalent to using
	 * the named constructor). Invalid names will throw an unchecked IllegalArgumentException.</p>
	 * 
	 * @param qualifiedName The fully qualified name of the type you wish to construct.
	 * @return The type, parented appropriately.
	 */
	public static JavaInterface interfaceFromQualifiedName(String qualifiedName) {
		if (!DataValidator.isValidPackageName(qualifiedName))
			throw new IllegalArgumentException("Qualified name must be valid Java identifier to construct type.");
		if (qualifiedName.indexOf(".") < 0) return new JavaInterface(qualifiedName);
		JavaPackage pkg = JavaPackage.fromString(qualifiedName.substring(0, qualifiedName.lastIndexOf(".")));
		JavaInterface result = new JavaInterface(qualifiedName.substring(qualifiedName.lastIndexOf(".")+1));
		result.setParent(pkg);
		
		return result;
	}
	
	public static JavaType getPrimitiveType(String primitive) {
		if (primitive.equals("boolean")) return BOOLEAN_TYPE;
		if (primitive.equals("byte")) return BYTE_TYPE;
		if (primitive.equals("char")) return CHAR_TYPE;
		if (primitive.equals("double")) return DOUBLE_TYPE;
		if (primitive.equals("float")) return FLOAT_TYPE;
		if (primitive.equals("int")) return INT_TYPE;
		if (primitive.equals("long")) return LONG_TYPE;
		if (primitive.equals("short")) return SHORT_TYPE;
		if (primitive.equals("void")) return VOID_TYPE;
		
		throw new IllegalArgumentException("Primitive argument must be a valid primitive keyword.");
	}
	
	public static final JavaType BOOLEAN_TYPE = new JavaClass("boolean");
	public static final JavaType BYTE_TYPE = new JavaClass("byte");
	public static final JavaType CHAR_TYPE = new JavaClass("char");
	public static final JavaType DOUBLE_TYPE = new JavaClass("double");
	public static final JavaType FLOAT_TYPE = new JavaClass("float");
	public static final JavaType INT_TYPE = new JavaClass("int");
	public static final JavaType LONG_TYPE = new JavaClass("long");
	public static final JavaType SHORT_TYPE = new JavaClass("short");
	public static final JavaType VOID_TYPE = new JavaClass("void");
	
	
}
